// 
// File:        EnumTest.java
// Copyright:   (c) 2001 The Regents of the University of California
// Release:     $Name: release-0-9-4 $
// Revision:    @(#) $Revision: 1.10 $
// Date:        $Date: 2004/07/02 02:22:40 $
// Description: enum regression test case for Java calling other languages
// 

/**
 * The following class runs the enum regression test cases for Java.
 */
public class EnumTest {
  private final static int FAIL        = 0;
  private final static int XFAIL       = 1;
  private final static int PASS        = 2;
  private final static int XPASS       = 3;
  private final static int UNSUPPORTED = 4;

  private final static String[] s_results = {
    "FAIL",
    "XFAIL",
    "PASS",
    "XPASS",
    "UNSUPPORTED"
  };

  private static int s_part = 0;
  private static int s_result = PASS;

  /**
   * Check the results of the test case.
   */
  private static void check(int expected, boolean pass, String test) {
    System.out.println("PART " + String.valueOf(++s_part));
    System.out.println("COMMENT " + test);
    System.out.print("RESULT " + String.valueOf(s_part) + " ");

    if (expected == UNSUPPORTED) {
      System.out.println(s_results[expected]);
      if (s_result == PASS) {
        s_result = XFAIL;
      }
    } else if (expected == PASS) {
      System.out.println(s_results[pass ? PASS : FAIL]);
      if (!pass) {
        s_result = FAIL;
      }
    } else if (expected == XFAIL) {
      System.out.println(s_results[pass ? XPASS : XFAIL]);
      if (pass) {
        s_result = FAIL;
      } else if (s_result > XFAIL) {
        s_result = XFAIL;
      }
    }
  }

  /**
   * The main test driver takes no command-line arguments and runs the
   * regression tests.
   */
  public static void main(String args[]) {
    try {
      /*
       * Begin the test
       */
      System.out.println("NPARTS -1");

      s_part   = 0;
      s_result = PASS;

      /*
       * Look at the pretty colors...
       */
      if (true) {
        sidl.Enum.Holder out   = new sidl.Enum.Holder(0);
        sidl.Enum.Holder inout = new sidl.Enum.Holder(enums.color.green);

        enums.colorwheel obj = new enums.colorwheel();

        check(PASS,
              (obj.returnback() == enums.color.violet),
             "(obj.returnback() == enums.color.violet)");
        check(PASS,
              (obj.passin(enums.color.blue) == true),
             "(obj.passin(enums.color.blue) == true)");
        check(PASS,
              (obj.passout(out) == true && out.get() == enums.color.violet),
             "(obj.passout(out) == true && out.get() == enums.color.violet)");
        check(PASS,
              (obj.passinout(inout) == true
            && inout.get() == enums.color.red),
             "(obj.passinout(inout) == true "
        + " && inout.get() == enums.color.red)");
        check(PASS,
              (obj.passeverywhere(enums.color.blue, out, inout) ==
                enums.color.violet 
            && out.get() == enums.color.violet
            && inout.get() == enums.color.green),
             "(obj.passeverywhere(enums.color.blue, out, inout) =="
        +     " enums.color.violet"
        + " && out.get() == enums.color.violet"
        + " && inout.get() == enums.color.green)");
      }

      /*
       * I want a porsche - zoom, zoom...
       */
      if (true) {
        sidl.Enum.Holder out   = new sidl.Enum.Holder(0);
        sidl.Enum.Holder inout = new sidl.Enum.Holder(enums.car.ford);

        enums.cartest obj = new enums.cartest();

        check(PASS,
              (obj.returnback() == enums.car.porsche),
             "(obj.returnback() == enums.car.porsche)");
        check(PASS,
              (obj.passin(enums.car.mercedes) == true),
             "(obj.passin(enums.car.mercedes) == true)");
        check(PASS,
              (obj.passout(out) == true && out.get() == enums.car.ford),
             "(obj.passout(out) == true && out.get() == enums.car.ford)");
        check(PASS,
              (obj.passinout(inout) == true
            && inout.get() == enums.car.porsche),
             "(obj.passinout(inout) == true "
        + " && inout.get() == enums.car.porsche)");
        check(PASS,
              (obj.passeverywhere(enums.car.mercedes, out, inout) ==
                enums.car.porsche
            && out.get() == enums.car.ford
            && inout.get() == enums.car.mercedes),
             "(obj.passeverywhere(enums.car.mercedes, out, inout) =="
        +     " enums.car.porsche"
        + " && out.get() == enums.car.ford"
        + " && inout.get() == enums.car.mercedes)");
      }

      /*
       * When I started programming, we only had zeros...
       */
      if (true) {
        sidl.Enum.Holder out   = new sidl.Enum.Holder(0);
        sidl.Enum.Holder inout = new sidl.Enum.Holder(enums.number.zero);

        enums.numbertest obj = new enums.numbertest();

        check(PASS,
              (obj.returnback() == enums.number.notOne),
             "(obj.returnback() == enums.number.notOne)");
        check(PASS,
              (obj.passin(enums.number.notZero) == true),
             "(obj.passin(enums.number.notZero) == true)");
        check(PASS,
              (obj.passout(out) == true && out.get() == enums.number.negOne),
             "(obj.passout(out) == true && out.get() == enums.number.negOne)");
        check(PASS,
              (obj.passinout(inout) == true
            && inout.get() == enums.number.notZero),
             "(obj.passinout(inout) == true "
        + " && inout.get() == enums.number.notZero)");
        check(PASS,
              (obj.passeverywhere(enums.number.notZero, out, inout) ==
                enums.number.notOne
            && out.get() == enums.number.negOne
            && inout.get() == enums.number.zero),
             "(obj.passeverywhere(enums.number.notZero, out, inout) =="
        +     " enums.number.notOne"
        + " && out.get() == enums.number.negOne"
        + " && inout.get() == enums.number.zero)");
      }
      
      if (true) {
	  int[] numElem = {2};
	  int[] stride = {2};
	  sidl.Enum.Array1 enumArray = new sidl.Enum.Array1(4, true);
	  sidl.Enum.Array1 slicedArray = null;
	  sidl.Enum.Array1 cpArray = null;
	  
	  check(PASS, enumArray != null, "enumArray != null");
	  enumArray.set(0, enums.car.porsche);
	  enumArray.set(1, enums.car.ford);
	  enumArray.set(2, enums.car.mercedes);
	  enumArray.set(3, enums.car.porsche);
	  check(PASS, enumArray.get(0) == enums.car.porsche,
		"Element 0 is a porsche!"); 
	  check(PASS, enumArray.get(1) == enums.car.ford,
		"Element 1 is a ford!");
	  check(PASS, enumArray.get(2) == enums.car.mercedes,
		"Element 2 is a mercedes!");
	  check(PASS, enumArray.get(3) == enums.car.porsche,
		"Element 3 is a porsche!");

       	  slicedArray = (sidl.Enum.Array1) enumArray._slice(1, numElem, (int[])null, stride, (int[])null)._dcast();
	  check(PASS, slicedArray != null, "Sliced Array is not null");
	  check(PASS, slicedArray.get(0) == enums.car.porsche,
	 	"Sliced Array Element 0 is a porche!"); 
	  check(PASS, slicedArray.get(1) == enums.car.mercedes,
		"Sliced Array Element 1 is a mercedes!");
	  cpArray = (sidl.Enum.Array1) slicedArray._smartCopy()._dcast();
	   check(PASS, cpArray != null, "Copied Array is not null");
	  check(PASS, cpArray._get(0) == enums.car.porsche,
	        "Copied Array Element 0 is a porche!"); 
	  check(PASS, cpArray._get(1) == enums.car.mercedes,
	        "Copied Array Element 1 is a mercedes!");
	  cpArray._deallocate();
	  slicedArray._deallocate();
	  enumArray._deallocate();
      

      }

      /*
       * Output final test results
       */

      System.out.println("TEST_RESULT " + s_results[s_result]);

      Runtime.getRuntime().exit(0); /* workaround for Linux JVM 1.3.1 bug */
    /*
     * Catch any unexpected exceptions and return a test failure
     */
    } catch (Throwable ex) {
      System.out.println("TEST_RESULT FAIL");
      System.out.println(ex.toString());
      System.exit(1);
    }
  }
}
